"""
Magic module that maps its submodules to Quilt tables.

Submodules have the following format: quilt.data.$user.$package.$table

E.g.:
  import quilt.data.$user.$package as $package
  print $package.$table
or
  from quilt.data.$user.$package import $table
  print $table

The corresponding data is looked up in `quilt_modules/$user/$package.h5`
in ancestors of the current directory.
"""

import imp
import os.path
import sys

from .tools.build import get_store
from .tools.store import PackageStore

__path__ = []  # Required for submodules to work

class Node(object):
    """
    Represents either the root of the store or a group, similar to nodes
    in HDFStore's `root`.
    """
    def __init__(self, store, prefix=''):
        self._prefix = prefix
        self._store = store

    def __getattr__(self, name):
        # TODO clean if... up since VALID_NAME_RE no longer allows leading _
        if name.startswith('_'):
            raise AttributeError
        path = self._prefix + '/' + name
        return self._get_store_obj(path)

    def __repr__(self):
        cinfo = str(self.__class__)
        finfo = 'File: ' + self._store.get_path()
        pinfo = 'Path: ' + self._prefix + '/'
        #TODO maybe show all descendant subpaths instead of just children
        spaths = [k + '/' for k in self._keys()]
        spaths.sort()
        output = [cinfo, finfo, pinfo] + spaths
        return '\n'.join(output)

    def _dfs(self):
        """
        every child key referencing a dataframe
        """
        pref = self._prefix + '/'
        return [k for k in self._keys()
                if not isinstance(self._get_store_obj(pref + k), Node)]

    def _get_store_obj(self, path):
        try:
            with self._store:
                return self._store.get(path)
        except KeyError:
            # No such group or table
            raise AttributeError("No such table or group: %s" % path)
        except TypeError:
            # This is awful, but that's what happens when the object being looked up
            # is a group rather than a table.
            return Node(self._store, path)

    def _groups(self):
        """
        every child key referencing a group that is not a dataframe
        """
        pref = self._prefix + '/'
        return [k for k in self._keys()
                if isinstance(self._get_store_obj(pref + k), Node)]

    def _keys(self):
        """
        keys directly accessible on this object via getattr or .
        """
        return self._store.keys(self._prefix)

class FakeLoader(object):
    """
    Fake module loader used to create intermediate user and package modules.
    """
    def __init__(self, path):
        self._path = path

    def load_module(self, fullname):
        """
        Returns an empty module.
        """
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = self._path
        mod.__loader__ = self
        mod.__path__ = []
        mod.__package__ = fullname
        return mod

class PackageLoader(object):
    """
    Module loader for Quilt tables.
    """
    def __init__(self, path, store):
        self._path = path
        self._store = store

    def load_module(self, fullname):
        """
        Returns an object that lazily looks up tables and groups.
        """
        mod = sys.modules.get(fullname)
        if mod is not None:
            return mod

        # We're creating an object rather than a module. It's a hack, but it's approved by Guido:
        # https://mail.python.org/pipermail/python-ideas/2012-May/014969.html

        mod = Node(self._store)
        sys.modules[fullname] = mod
        return mod

class ModuleFinder(object):
    """
    Looks up submodules.
    """
    @staticmethod
    def find_module(fullname, path=None):
        """
        Looks up the table based on the module path.
        """
        if not fullname.startswith(__name__ + '.'):
            # Not a quilt submodule.
            return None

        submodule = fullname[len(__name__) + 1:]
        parts = submodule.split('.')

        if len(parts) == 1:
            for package_dir in PackageStore.find_package_dirs():
                file_path = os.path.join(package_dir, parts[0])
                if os.path.isdir(file_path):
                    return FakeLoader(file_path)
        elif len(parts) == 2:
            user, package = parts
            store = get_store(user, package)
            if store:
                file_path = store.get_path()
                return PackageLoader(file_path, store)

        return None

sys.meta_path.append(ModuleFinder)



import os

import yaml
import pandas as pd
from .store import get_store, VALID_NAME_RE, StoreException
from .const import TARGET

class BuildException(Exception):
    """
    Build-time exception class
    """
    pass

def _build_table(build_dir, store, name, table, target='pandas'):
    if isinstance(table, list):
        if len(table) != 2:
            raise BuildException(
                "Table definition must be a list of [type, path]")
        ext, rel_path = table
        path = os.path.join(build_dir, rel_path)
        # read source file into DataFrame
        df = _file_to_data_frame(ext, path, target)
        # serialize DataFrame to file(s)
        store.save_df(df, name, path, ext, target)

    elif isinstance(table, dict):
        # TODO the problem with this, it does not seem to iterate
        # in the same order as the entries in the file, which is
        # weird as users might expect error/success messages to be in
        # file order
        for child_name, child_table in table.items():
            if not isinstance(child_name, str) or not VALID_NAME_RE.match(child_name):
                raise StoreException("Invalid table name: %r" % child_name)
            _build_table(build_dir, store, name + '/' + child_name, child_table)
    else:
        raise BuildException("Table definition must be a list or dict")

def _file_to_data_frame(ext, path, target):
    ext = ext.lower() #ensure that case doesn't matter
    platform = TARGET.get(target)
    if platform is None:
        raise BuildException('Unsupported target platform: %s' % target)
    logic = platform.get(ext)
    if logic is None:
        raise BuildException('Unsupported input file type: .%s' % ext)
    fname = logic['attr']
    kwargs = logic['kwargs']
    failover = logic.get('failover', None)
    handler = getattr(pd, fname, None)
    if handler is None:
        raise BuildException("Invalid ingest function: %r" % fname)

    df = None
    try_again = False
    try:
        df = handler(path, **kwargs)
    except UnicodeDecodeError as error:
        if failover:
            warning = "Warning: failed fast parse on input %s.\n" % path
            warning += "Switching to Python engine."
            print(warning)
            try_again = True
        else:
            raise error

    if try_again:
        failover_args = {}
        failover_args.update(failover)
        failover_args.update(kwargs)
        df = handler(path, **failover_args)

    return df

def build_package(username, package, yaml_path):
    """
    Builds a package from a given Yaml file and installs it locally.

    Returns the name of the package.
    """
    try:
        build_dir = os.path.dirname(yaml_path)

        fd = open(yaml_path)
        docs = yaml.load_all(fd)
        data = next(docs, None) # leave other dicts in the generator
        if not isinstance(data, dict):
            raise BuildException("Unable to parse YAML: %s" % yaml_path)

        tables = data.get('tables')
        format = data.get('format', None)
        if not isinstance(tables, dict):
            raise BuildException("'tables' must be a dictionary")

        with get_store(username, package, format, 'w') as store:
            _build_table(build_dir, store, '', tables)
    except (IOError, StoreException) as ex:
        raise BuildException(str(ex))



"""
Test the build process
"""
#TODO: we should really test the CLI interface itself, rather than
#the functions that cli calls
import os
import sys

import pytest

from quilt.tools import build
from quilt.tools.const import FORMAT_PARQ

try:
    import fastparquet
except ImportError:
    fastparquet = None

USER = 'test_121'
PACKAGE = 'groot'

def test_build():
    """
    Test compilation
    """
    mydir = os.path.dirname(__file__)
    PATH = os.path.join(mydir, './build.yml')
    build.build_package(USER, PACKAGE, PATH)
    # TODO load DFs based on contents of .yml file at PATH
    # not hardcoded vals (this will require loading modules from variable
    # names, probably using __module__)
    from quilt.data.test_121.groot import csv, tsv, xls
    rows = len(csv.index)
    assert rows == len(tsv.index) and rows == len(xls.index), \
        'Expected dataframes to have same # rows'
    cols = len(csv.columns)
    print(csv.columns, xls.columns, tsv.columns)
    assert cols == len(tsv.columns) and cols == len(xls.columns), \
        'Expected dataframes to have same # columns'
    # TODO add more integrity checks, incl. negative test cases

@pytest.mark.skipif("fastparquet is None")
def test_build_parquet():
    """
    Test compilation
    """
    os.environ["QUILT_PACKAGE_FORMAT"] = FORMAT_PARQ
    mydir = os.path.dirname(__file__)
    PATH = os.path.join(mydir, './build.yml')
    build.build_package(USER, PACKAGE, PATH)
    # TODO load DFs based on contents of .yml file at PATH
    # not hardcoded vals (this will require loading modules from variable
    # names, probably using __module__)
    from quilt.data.test_121.groot import csv, tsv, xls
    rows = len(csv.index)
    assert rows == len(tsv.index) and rows == len(xls.index), \
        'Expected dataframes to have same # rows'
    cols = len(csv.columns)
    print(csv.columns, xls.columns, tsv.columns)
    assert cols == len(tsv.columns) and cols == len(xls.columns), \
        'Expected dataframes to have same # columns'
    del(os.environ["QUILT_PACKAGE_FORMAT"])
    # TODO add more integrity checks, incl. negative test cases




"""
Build: parse and add user-supplied files to store
"""
from operator import attrgetter
import os
import re
import tempfile
import time
import zlib

import pandas as pd
import requests

try:
    import fastparquet
except ImportError:
    fastparquet = None

from .const import DTIMEF, FORMAT_HDF5, FORMAT_PARQ
from .hashing import digest_file

# start with alpha (_ may clobber attrs), continue with alphanumeric or _
VALID_NAME_RE = re.compile(r'^[a-zA-Z]\w*$')
CHUNK_SIZE = 4096
ZLIB_LEVEL = 2  # Maximum level.
ZLIB_METHOD = zlib.DEFLATED  # The only supported one.
ZLIB_WBITS = zlib.MAX_WBITS | 16  # Add a gzip header and checksum.


class StoreException(Exception):
    """
    Exception class for store I/O
    """
    pass


class PackageStore(object):
    """
    Base class for managing Quilt data package repositories. This
    class and its subclasses abstract file formats, file naming and
    reading and writing to/from data files.
    """
    PACKAGE_DIR_NAME = 'quilt_packages'

    @classmethod
    def find_package_dirs(cls, start='.'):
        """
        Walks up the directory tree and looks for `quilt_packages` directories
        in the ancestors of the starting directory.

        The algorithm is the same as Node's `node_modules` algorithm
        ( https://nodejs.org/docs/v7.4.0/api/modules.html#modules_all_together ),
        except that it doesn't stop at the top-level `quilt_packages` directory.

        Returns a (possibly empty) generator.
        """
        path = os.path.realpath(start)
        while True:
            parent_path, name = os.path.split(path)
            if name != cls.PACKAGE_DIR_NAME:
                package_dir = os.path.join(path, cls.PACKAGE_DIR_NAME)
                if os.path.isdir(package_dir):
                    yield package_dir
            if parent_path == path:  # The only reliable way to detect the root.
                break
            path = parent_path

    def __init__(self, user, package, mode):
        self._user = user
        self._package = package
        self._mode = mode
        self._path = self.get_path()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass

    def keys(self, prefix):
        """
        Returns a list of package contents.
        """
        raise StoreException("Not Implemented")

    def get_hash(self):
        """
        Returns the hash digest of the package data.
        """
        raise StoreException("Not Implemented")

    def create_path(self):
        """
        Creates a path to store a data package in the innermost `quilt_packages`
        directory (or in a new `quilt_packages` directory in the current directory)
        and allocates a per-user directory if needed.
        """
        if not VALID_NAME_RE.match(self._user):
            raise StoreException("Invalid user name: %r" % self._user)
        if not VALID_NAME_RE.match(self._package):
            raise StoreException("Invalid package name: %r" % self._package)

        package_dir = next(PackageStore.find_package_dirs(), self.PACKAGE_DIR_NAME)
        user_path = os.path.join(package_dir, self._user)
        if not os.path.isdir(user_path):
            os.makedirs(user_path)
        path = os.path.join(user_path, self._package)
        return path

    def get_path(self):
        """
        Finds an existing package in one of the package directories.
        """
        if not VALID_NAME_RE.match(self._user):
            raise StoreException("Invalid user name: %r" % self._user)
        if not VALID_NAME_RE.match(self._package):
            raise StoreException("Invalid package name: %r" % self._package)

        pkg_dirs = PackageStore.find_package_dirs()
        for package_dir in pkg_dirs:
            path = os.path.join(package_dir, self._user, self._package)
            if os.path.exists(path):
                return path
        return None


class HDF5PackageStore(PackageStore):
    """
    HDF5 Implementation of PackageStore.
    """
    PACKAGE_FILE_EXT = '.h5'

    def __init__(self, user, package, mode):
        super(HDF5PackageStore, self).__init__(user, package, mode)
        self.__store = None

    def __enter__(self):
        if self._mode == 'w' and self._path is None:
            self._path = self.create_path()
        self.__get_store()
        return self

    def __exit__(self, type, value, traceback):
        self.__store.close()
        self.__store = None

    def __get_store(self):
        if self.__store is None:
            self.__store = pd.HDFStore(self._path, mode=self._mode)
        return self.__store

    def exists(self):
        """
        Returns True if the package is already installed.
        """
        return not self._path is None

    def create_path(self):
        """
        Creates a new .h5 file in the innermost `quilt_packages` directory
        (or in a new `quilt_packages` directory in the current directory).
        Overwrites any existing package.
        """
        path = super(HDF5PackageStore, self).create_path() + self.PACKAGE_FILE_EXT
        return path

    def get_path(self):
        """
        Finds an existing package in one of the package directories.
        """
        if not VALID_NAME_RE.match(self._user):
            raise StoreException("Invalid user name: %r" % self._user)
        if not VALID_NAME_RE.match(self._package):
            raise StoreException("Invalid package name: %r" % self._package)

        pkg_dirs = PackageStore.find_package_dirs()
        for package_dir in pkg_dirs:
            path = os.path.join(package_dir, self._user, self._package + self.PACKAGE_FILE_EXT)
            if os.path.exists(path):
                return path
        return None

    def get(self, path):
        """
        Read a DataFrame from the store.
        """
        if not self.exists():
            raise StoreException("Package not found")

        return self.__get_store().get(path)

    def get_hash(self):
        path = self.get_path()
        if path is None:
            raise StoreException("Package not found")
        return digest_file(path)

    class UploadFile(object):
        """
        Helper class to manage temporary package files uploaded by push.
        """
        def __init__(self, store):
            self._store = store
            self._path = store.get_path()

        def __enter__(self):
            self._temp_file = tempfile.TemporaryFile()
            with open(self._path, 'rb') as input_file:
                zlib_obj = zlib.compressobj(ZLIB_LEVEL, ZLIB_METHOD, ZLIB_WBITS)
                for chunk in iter(lambda: input_file.read(CHUNK_SIZE), b''):
                    self._temp_file.write(zlib_obj.compress(chunk))
                self._temp_file.write(zlib_obj.flush())
            self._temp_file.seek(0)
            return self._temp_file

        def __exit__(self, type, value, traceback):
            self._temp_file.close()

    def tempfile(self):
        """
        Create and return a temporary file for uploading to a registry.
        """
        return self.UploadFile(self)

    def install(self, url, download_hash):
        """
        Download and install a package locally.
        """
        local_filename = self.create_path()

        response = requests.get(url, stream=True)
        if not response.ok:
            raise StoreException("Download failed: error %s" % response.status_code)

        assert local_filename, "Blank filename? %s" % local_filename

        with open(local_filename, 'wb') as output_file:
            # `requests` will automatically un-gzip the content, as long as
            # the 'Content-Encoding: gzip' header is set.
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                if chunk: # filter out keep-alive new chunks
                    output_file.write(chunk)

        file_hash = self.get_hash()
        if file_hash != download_hash:
            os.remove(local_filename)
            raise StoreException("Mismatched hash! Expected %s, got %s." %
                                 (download_hash, file_hash))


    def keys(self, prefix):
        # prepending root ensures dots is never empty, preventing an
        # exception from attrgetter
        dots = 'root' + prefix.replace('/', '.')
        node = attrgetter(dots)(self.__get_store())
        return node._v_children.keys()

    def save_df(self, df, name, path, ext, target):
        """
        Save a DataFrame to the store.
        """
        self.__store[name] = df

        # add metadata as HDF5 attrs
        dots = 'root' + name.replace('/', '.')
        snode = attrgetter(dots)(self.__store)
        # in spite of pytables docs
        # http://www.pytables.org/usersguide/libref/declarative_classes.html#attributesetclassdescr
        # snode.attrs does not work
        snode._v_attrs.q_ext = ext
        snode._v_attrs.q_path = path
        snode._v_attrs.q_target = target
        snode._v_attrs.q_timestamp = time.strftime(DTIMEF, time.gmtime())

    @classmethod
    def ls_packages(cls, pkg_dir):
        """
        List installed packages.
        """
        hdf5_packages = [
            (user, pkg[:-len(HDF5PackageStore.PACKAGE_FILE_EXT)])
            for user in os.listdir(pkg_dir)
            for pkg in os.listdir(os.path.join(pkg_dir, user))
            if pkg.endswith(HDF5PackageStore.PACKAGE_FILE_EXT)]
        return hdf5_packages


class ParquetPackageStore(PackageStore):
    """
    Parquet Implementation of PackageStore.
    """

    PACKAGE_FILE_EXT = '.parq'

    def __init__(self, user, package, mode):
        super(ParquetPackageStore, self).__init__(user, package, mode)
        if fastparquet is None:
            raise StoreException("Module fastparquet is required for ParquetPackageStore.")

        if self._mode == 'w':
            path = self.create_path()
        else:
            path = self.get_path()
        self.active_path = path

    def create_path(self):
        """
        Creates a new subdirectory in the innermost `quilt_packages` directory
        (or in a new `quilt_packages` directory in the current directory).
        """
        path = super(ParquetPackageStore, self).create_path()
        if not os.path.isdir(path):
            os.makedirs(path)
        return path

    def save_df(self, df, name, path, ext, target):
        """
        Save a DataFrame to the store.
        """
        # Below should really use os.path.join, but name is
        # arriving with a leading / that breaks it.
        path = self.active_path + name + self.PACKAGE_FILE_EXT
        fastparquet.write(path, df)

    def get(self, path):
        """
        Read a DataFrame to the store.
        """
        fpath = self.get_path() + path + self.PACKAGE_FILE_EXT
        pfile = fastparquet.ParquetFile(fpath)
        return pfile.to_pandas()

    def get_hash(self):
        raise StoreException("Not Implemented")

    @classmethod
    def ls_packages(cls, pkg_dir):
        """
        List installed packages.
        """
        parq_packages = [
            (user, pkg)
            for user in os.listdir(pkg_dir)
            for pkg in os.listdir(os.path.join(pkg_dir, user))
            if os.path.isdir(pkg)]
        return parq_packages


# Helper functions
def get_store(user, package, format=None, mode='r'):
    """
    Return a PackageStore object of the appropriate type for a
    given data package.
    """
    pkg_format = format
    if not pkg_format:
        pkg_format = os.environ.get('QUILT_PACKAGE_FORMAT', FORMAT_HDF5)
    if pkg_format == FORMAT_PARQ:
        return ParquetPackageStore(user, package, mode)
    else:
        return HDF5PackageStore(user, package, mode)

def ls_packages(pkg_dir):
    """
    List all packages from all package directories.
    """
    pkg_format = os.environ.get('QUILT_PACKAGE_FORMAT', FORMAT_HDF5)
    if pkg_format == FORMAT_HDF5:
        packages = HDF5PackageStore.ls_packages(pkg_dir)
    elif pkg_format == FORMAT_PARQ:
        packages = ParquetPackageStore.ls_packages(pkg_dir)
    else:
        raise StoreException("Unsupported Package Format %s" % pkg_format)
    return packages



"""
Helper functions.
"""
import os.path
import re

from appdirs import user_data_dir

APP_NAME = "QuiltCli"
APP_AUTHOR = "QuiltData"
BASE_DIR = user_data_dir(APP_NAME, APP_AUTHOR)

def file_to_str(fname):
    """
    Read a file into a string
    PRE: fname is a small file (to avoid hogging memory and its discontents)
    """
    data = None
    # rU = read with Universal line terminator
    with open(fname, 'rU') as f:
        data = f.read()
    return data



"""
Tests for commands.
"""

from argparse import Namespace
from unittest import TestCase
try:
    # Python3
    from unittest.mock import patch
except ImportError:
    # Python2 - external dependency.
    from mock import patch

import pytest
import responses

try:
    import h5py
except ImportError:
    h5py = None

from quilt.tools import command

class CommandTest(TestCase):
    # Note: we're using the deprecated `assertRaisesRegexp` method because
    # the new one, `assertRaisesRegex`, is not present in Python2.

    def test_push_invalid_package(self):
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.push(Namespace(package="no_user"))
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.push(Namespace(package="a/b/c"))

    def test_install_invalid_package(self):
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.install(Namespace(package="no_user"))
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.install(Namespace(package="a/b/c"))

    @pytest.mark.skipif("h5py is None")
    def test_inspect_invalid_package(self):
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.inspect(Namespace(package="no_user"))
        with self.assertRaisesRegexp(command.CommandException, "owner/package_name"):
            command.inspect(Namespace(package="a/b/c"))

    def test_push_missing_package(self):
        with self.assertRaisesRegexp(command.CommandException, "not found"):
            command.push(Namespace(package="owner/package"))

    @pytest.mark.skipif("h5py is None")
    def test_inspect_missing_package(self):
        with self.assertRaisesRegexp(command.CommandException, "not found"):
            command.inspect(Namespace(package="owner/package"))

    @responses.activate
    @patch('webbrowser.open')
    @patch('quilt.tools.command.input')
    @patch('quilt.tools.command._save_auth')
    def test_login(self, mock_save, mock_input, mock_open):
        mock_input.return_value = "123"

        responses.add(
            responses.POST,
            '%s/api/token' % command.QUILT_PKG_URL,
            json=dict(
                statuc=200,
                refresh_token="456",
                access_token="abc",
                expires_at=1000.0
            )
        )

        command.login(Namespace())

        mock_open.assert_called_with('%s/login' % command.QUILT_PKG_URL)

        assert responses.calls[0].request.body == "refresh_token=123"

        mock_save.assert_called_with(dict(
            refresh_token="456",
            access_token="abc",
            expires_at=1000.0
        ))

    @responses.activate
    @patch('webbrowser.open')
    @patch('quilt.tools.command.input')
    @patch('quilt.tools.command._save_auth')
    def test_login_server_error(self, mock_save, mock_input, mock_open):
        mock_input.return_value = "123"

        responses.add(
            responses.POST,
            '%s/api/token' % command.QUILT_PKG_URL,
            status=500
        )

        with self.assertRaises(command.CommandException):
            command.login(Namespace())

        mock_save.assert_not_called()

    @responses.activate
    @patch('webbrowser.open')
    @patch('quilt.tools.command.input')
    @patch('quilt.tools.command._save_auth')
    def test_login_auth_fail(self, mock_save, mock_input, mock_open):
        mock_input.return_value = "123"

        responses.add(
            responses.POST,
            '%s/api/token' % command.QUILT_PKG_URL,
            json=dict(
                status=200,
                error="Bad token!"
            )
        )

        with self.assertRaises(command.CommandException):
            command.login(Namespace())

        mock_save.assert_not_called()




# -*- coding: utf-8 -*-

from __future__ import print_function
from builtins import input
import argparse
import json
import os
import stat
import sys
import time
import webbrowser

import requests

from .build import build_package, BuildException
from .const import LATEST_TAG
from .store import PackageStore, StoreException, get_store, ls_packages
from .util import BASE_DIR

HEADERS = {"Content-Type": "application/json", "Accept": "application/json"}

QUILT_AUTH_URL = os.environ.get('QUILT_AUTH_URL', 'https://quiltdata.com')
QUILT_PKG_URL = os.environ.get('QUILT_PKG_URL', 'https://pkg.quiltdata.com')

AUTH_FILE_NAME = "auth.json"


class CommandException(Exception):
    """
    Exception class for all command-related failures.
    """
    pass


def _update_auth(refresh_token):
    response = requests.post("%s/api/token" % QUILT_PKG_URL, data=dict(
        refresh_token=refresh_token
    ))

    if response.status_code != requests.codes.ok:
        raise CommandException("Authentication error: %s" % response.status_code)

    data = response.json()
    error = data.get('error')
    if error is not None:
        raise CommandException("Failed to log in: %s" % error)

    return dict(
        refresh_token=data['refresh_token'],
        access_token=data['access_token'],
        expires_at=data['expires_at']
    )

def _save_auth(auth):
    if not os.path.exists(BASE_DIR):
        os.makedirs(BASE_DIR)

    file_path = os.path.join(BASE_DIR, AUTH_FILE_NAME)
    with open(file_path, 'w') as fd:
        os.chmod(file_path, stat.S_IRUSR | stat.S_IWUSR)
        json.dump(auth, fd)

def _create_session():
    try:
        file_path = os.path.join(BASE_DIR, AUTH_FILE_NAME)
        with open(file_path) as fd:
            auth = json.load(fd)
    except (IOError, ValueError):
        raise CommandException("Please run `quilt login` first.")

    # If the access token expires within a minute, update it.
    if auth['expires_at'] < time.time() + 60:
        try:
            auth = _update_auth(auth['refresh_token'])
        except CommandException as ex:
            raise CommandException(
                "Failed to update the access token (%s). Run `quilt login` again." % ex
            )
        _save_auth(auth)

    session = requests.Session()
    session.headers.update({
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "Bearer %s" % auth['access_token']
    })

    response = session.get("%s/api-root" % QUILT_AUTH_URL)
    if response.status_code == requests.codes.unauthorized:
        raise CommandException("Failed to log in. Run `quilt login` again.")
    elif response.status_code != requests.codes.ok:
        raise CommandException("Failed to access Quilt.")

    return session

def _parse_package(name):
    try:
        owner, pkg = name.split('/')
        if not owner or not pkg:
            # Make sure they're not empty.
            raise ValueError
    except ValueError:
        raise CommandException("Specify package as owner/package_name.")
    return owner, pkg

def login(args):
    """
    Authenticate.
    """
    # Open the browser. Get rid of stdout while launching the browser to prevent
    # Chrome/Firefox from outputing garbage over the code prompt.
    devnull = os.open(os.devnull, os.O_RDWR)
    old_stdout = os.dup(1)
    os.dup2(devnull, 1)
    try:
        webbrowser.open("%s/login" % QUILT_PKG_URL)
    finally:
        os.close(devnull)
        os.dup2(old_stdout, 1)
        os.close(old_stdout)

    refresh_token = input("Enter the code: ")

    # Get an access token (and a new refresh token).
    # Technically, we could have the user enter both tokens - but it doesn't
    # really matter, and this lets us verify that the token actually works.
    auth = _update_auth(refresh_token)

    _save_auth(auth)

def build(args):
    """
    Compile a Quilt data package
    """
    owner, pkg = _parse_package(args.package)
    try:
        build_package(owner, pkg, args.path)
        print("Built %s/%s successfully." % (owner, pkg))
    except BuildException as ex:
        raise CommandException("Failed to build the package: %s" % ex)

def push(args):
    """
    Push a Quilt data package to the server
    """
    owner, pkg = _parse_package(args.package)

    store = get_store(owner, pkg)
    if not store.exists():
        raise CommandException("Package {owner}/{pkg} not found.".format(owner=owner, pkg=pkg))

    path = store.get_path()
    pkghash = store.get_hash()
    assert path

    session = _create_session()

    response = session.put(
        "{url}/api/package/{owner}/{pkg}/{hash}".format(
            url=QUILT_PKG_URL,
            owner=owner,
            pkg=pkg,
            hash=pkghash
        ),
        data=json.dumps(dict(
            description=""  # TODO
        ))
    )

    if not response.ok:
        # PUT Dataset failed
        raise CommandException("Push failed: error %s" % response.status_code)

    dataset = response.json()
    upload_url = dataset['upload_url']

    headers = {
        'Content-Encoding': 'gzip'
    }

    # Create a temporary gzip'ed file.
    with store.tempfile() as temp_file:
        response = requests.put(upload_url, data=temp_file, headers=headers)

        if not response.ok:
            raise CommandException("Upload failed: error %s" % response.status_code)

    # Set the "latest" tag.
    response = session.put(
        "{url}/api/tag/{owner}/{pkg}/{tag}".format(
            url=QUILT_PKG_URL,
            owner=owner,
            pkg=pkg,
            tag=LATEST_TAG
        ),
        data=json.dumps(dict(
            hash=pkghash
        ))
    )
    if not response.ok:
        raise CommandException("Failed to set the 'latest' tag: error %s" % response.status_code)

def install(args):
    """
    Download a Quilt data package from the server and install locally
    """
    owner, pkg = _parse_package(args.package)
    store = get_store(owner, pkg, mode='w')

    if store.exists():
        print("{owner}/{pkg} already installed.".format(owner=owner, pkg=pkg))
        overwrite = input("Overwrite y/n? ")
        if overwrite.lower() != 'y':
            return

    session = _create_session()

    # Get the "latest" tag.
    response = session.get(
        "{url}/api/tag/{owner}/{pkg}/{tag}".format(
            url=QUILT_PKG_URL,
            owner=owner,
            pkg=pkg,
            tag=LATEST_TAG
        )
    )
    if response.status_code == requests.codes.not_found:
        raise CommandException("Lookup failed: package {owner}/{pkg} not found.".format(
            owner=owner, pkg=pkg))
    elif not response.ok:
        raise CommandException("Failed to get the 'latest' tag: error %s" % response.status_code)

    pkghash = response.json()['hash']

    response = session.get(
        "{url}/api/package/{owner}/{pkg}/{hash}".format(
            url=QUILT_PKG_URL,
            owner=owner,
            pkg=pkg,
            hash=pkghash
        )
    )
    if not response.ok:
        raise CommandException("Lookup failed: error %s" % response.status_code)
    dataset = response.json()

    store.install(dataset['url'], dataset['hash'])

def access_list(args):
    """
    Print list of users who can access a package.
    """
    owner, pkg = _parse_package(args.package)

    session = _create_session()

    lookup_url = "{url}/api/access/{owner}/{pkg}".format(url=QUILT_PKG_URL, owner=owner, pkg=pkg)
    response = session.get(lookup_url)
    if response.status_code == requests.codes.not_found:
        raise CommandException("Lookup failed: package {owner}/{pkg} not found.".format(
            owner=owner, pkg=pkg))
    elif not response.ok:
        raise CommandException("Lookup failed: error %s" % response.status_code)

    data = response.json()
    users = data['users']

    print('\n'.join(users))

def access_add(args):
    owner, pkg = _parse_package(args.package)
    user = args.user

    session = _create_session()

    response = session.put("%s/api/access/%s/%s/%s" % (QUILT_PKG_URL, owner, pkg, user))
    if response.status_code == requests.codes.not_found:
        raise CommandException("Failed to add access: package {owner}/{pkg} not found.".format(
            owner=owner, pkg=pkg))
    elif not response.ok:
        raise CommandException("Failed to add access: %s" % response.status_code)

def access_remove(args):
    owner, pkg = _parse_package(args.package)
    user = args.user

    session = _create_session()

    response = session.delete("%s/api/access/%s/%s/%s" % (QUILT_PKG_URL, owner, pkg, user))
    if response.status_code == requests.codes.not_found:
        raise CommandException("Failed to remove access: package {owner}/{pkg} not found.".format(
            owner=owner, pkg=pkg))
    elif not response.ok:
        raise CommandException("Failed to remove access: %s" % response.status_code)

def ls(args):
    """
    List all installed Quilt data packages
    """
    for pkg_dir in PackageStore.find_package_dirs():
        print("%s" % pkg_dir)
        packages = ls_packages(pkg_dir)
        for idx, (owner, pkg) in enumerate(packages):
            prefix = u"└── " if idx == len(packages) - 1 else u"├── "
            print("%s%s/%s" % (prefix, owner, pkg))

def inspect(args):
    """
    Inspect package details
    """
    try:
        import h5py
    except ImportError:
        raise CommandException("Please install 'h5py' to use 'quilt inspect'")

    owner, pkg = _parse_package(args.package)
    store = get_store(owner, pkg)
    if not store.exists():
        raise CommandException("Package {owner}/{pkg} not found.".format(owner=owner, pkg=pkg))
    path = store.get_path()
    assert path

    def _print_children(children, prefix):
        for idx, child in enumerate(children):
            if idx == len(children) - 1:
                new_prefix = u"└─"
                new_child_prefix = u"  "
            else:
                new_prefix = u"├─"
                new_child_prefix = u"│ "
            _print_node(child, prefix + new_prefix, prefix + new_child_prefix)

    def _print_node(node, prefix, child_prefix):
        name_prefix = u"─ "
        if isinstance(node, h5py.Group):
            children = list(node.values())
            if children:
                name_prefix = u"┬ "
            print(prefix + name_prefix + node.name)
            _print_children(children, child_prefix)
        elif isinstance(node, h5py.Dataset):
            info = "shape %s, type \"%s\"" % (node.shape, node.dtype.str)
            print(prefix + name_prefix + node.name + ": " + info)
        else:
            print(prefix + name_prefix + node.name + ": " + str(node))

    h5_file = h5py.File(path)
    print(path)
    _print_children(list(h5_file.values()), '')

def main():
    parser = argparse.ArgumentParser(description="Quilt Command Line")
    subparsers = parser.add_subparsers(title="Commands", dest='cmd')
    subparsers.required = True

    login_p = subparsers.add_parser("login")
    login_p.set_defaults(func=login)

    build_p = subparsers.add_parser("build")
    build_p.add_argument("package", type=str, help="Owner/Package Name")
    build_p.add_argument("path", type=str, help="Path to the Yaml build file")
    build_p.set_defaults(func=build)

    push_p = subparsers.add_parser("push")
    push_p.add_argument("package", type=str, help="Owner/Package Name")
    push_p.set_defaults(func=push)

    install_p = subparsers.add_parser("install")
    install_p.add_argument("package", type=str, help="Owner/Package Name")
    install_p.set_defaults(func=install)

    access_p = subparsers.add_parser("access")
    access_subparsers = access_p.add_subparsers(title="Access", dest='cmd')
    access_subparsers.required = True

    access_list_p = access_subparsers.add_parser("list")
    access_list_p.add_argument("package", type=str, help="Owner/Package Name")
    access_list_p.set_defaults(func=access_list)

    access_add_p = access_subparsers.add_parser("add")
    access_add_p.add_argument("package", type=str, help="Owner/Package Name")
    access_add_p.add_argument("user", type=str, help="User to add")
    access_add_p.set_defaults(func=access_add)

    access_remove_p = access_subparsers.add_parser("remove")
    access_remove_p.add_argument("package", type=str, help="Owner/Package Name")
    access_remove_p.add_argument("user", type=str, help="User to remove")
    access_remove_p.set_defaults(func=access_remove)

    ls_p = subparsers.add_parser("ls")
    ls_p.set_defaults(func=ls)

    inspect_p = subparsers.add_parser("inspect")
    inspect_p.add_argument("package", type=str, help="Owner/Package Name")
    inspect_p.set_defaults(func=inspect)

    args = parser.parse_args()
    try:
        args.func(args)
        return 0
    except CommandException as ex:
        print(ex, file=sys.stderr)
        return 1



"""
Constants
"""
DATEF = '%F'
TIMEF = '%T'
DTIMEF = '%s %s' % (DATEF, TIMEF)

LATEST_TAG = 'latest'

# SHA-2 Family
HASH_TYPE = 'sha256'
# RSA digital signature key Size
RSA_BITS = 2048

FORMAT_HDF5 = 'HDF5'
FORMAT_PARQ = 'Parquet'

# TODO nan probably not a safe choice and may pollute number cols with strs
NA_VALS = ['nan']
KEEP_NA = False
# Supported build targets and file types
# BUILD[target][file_extension]
# file_extension should be lowercase
TARGET = {
    'pandas': {
        'csv': {
            'attr': 'read_csv',
            'failover' : {'engine' : 'python'},
            'kwargs': {
                'keep_default_na': KEEP_NA,
                'na_values': NA_VALS,
            }
        },
        'ssv': {
            'attr': 'read_csv',
            'failover' : {'engine' : 'python'},
            'kwargs': {
                'keep_default_na': KEEP_NA,
                'na_values': NA_VALS,
                'sep': ';'
            }
        },
        'tsv': {
            'attr': 'read_csv',
            'failover' : {'engine' : 'python'},
            'kwargs': {
                'keep_default_na': KEEP_NA,
                'na_values': NA_VALS,
                'sep': '\t'
            }
        },
        'xls': {
            'attr': 'read_excel',
            # TODO set sheetname='None' to get all sheets?
            # Currently defaults to sheetname=0, which imports first sheet only
            'kwargs': {
                'keep_default_na': KEEP_NA,
                'na_values': NA_VALS
            }
        },
        'xlsx': {
            'attr': 'read_excel',
            # see comments under 'xls'
            'kwargs': {
                'keep_default_na': KEEP_NA,
                'na_values': NA_VALS
            }
        }
    }
}



from setuptools import setup, find_packages

def readme():
    with open('README.md', 'r') as f:
        return f.read()

setup(
    name="quilt",
    version="2.0.0",
    packages=find_packages(),
    description='Quilt is an open-source data frame registry',
    long_description=readme(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
    ],
    author='quiltdata',
    author_email='founders@quiltdata.io',
    license='LICENSE',
    url='https://github.com/quiltdata/quilt',
    download_url='https://github.com/quiltdata/quilt/releases/tag/v2.0.0-alpha',
    keywords='quilt quiltdata shareable data dataframe package platform pandas',
    install_requires=[
        'appdirs>=1.4.0',
        'future>=0.16.0',
        'pandas>=0.19.2',
        'pyOpenSSL>=16.2.0',
        'pyyaml>=3.12',
        'requests>=2.12.4',
        'responses>=0.5.1',
        'tables>=3.3.0',
        'xlrd>=1.0.0',
    ],
    entry_points={
        'console_scripts': ['quilt=quilt.tools.command:main'],
    }
)
